//
//  HexagonTiledTorus.swift
//  HexagonTiledTorus
//
//  Created by Don McBrien on 20/02/2020.
//  Copyright © 2020 Don McBrien. All rights reserved.
//

import Foundation

public class HexagonTiledTorus: CustomStringConvertible {
   
   //MARK: Constants
   // Set by init()
   fileprivate let meridians: Int
   fileprivate let ½ParallelsPerHemisphere: Int
   fileprivate let exquatorRadius: Float
   
   // Calculated up-front by setConstants()
   fileprivate var ½AspectRatio: Float = sqrt(3.0) / 4.0
   fileprivate var distortion: Float!
   fileprivate var ¾HexWidth: Angle!        // ¾ angular width
   fileprivate var ½HexWidth: Angle!        // ½ angular width
   
   fileprivate var tubeRadius: Float!
   fileprivate var rimRadius: Float!
   fileprivate var radiusRatio: Float!

   fileprivate var yellow = SIMD4<Float>(0.6,0.2,0.2,1)
   fileprivate var red = SIMD4<Float>(1,0.3,0.3,1)
   
   //MARK: Variables
   public var vertices = [Vertex]()
   public var indices = [Triangle]()

   fileprivate var parallels = [Parallels]()
   fileprivate var hexagons = [Int:Hexagon]()
   
   private func setConstants() {
      distortion = ½AspectRatio * 4.0 / Float(meridians)
      
      tubeRadius = exquatorRadius * (1.0 - pow((1.0 - distortion), Float(½ParallelsPerHemisphere))) / 2.0
      rimRadius = exquatorRadius - tubeRadius

      ¾HexWidth = Angle(inDegrees: 360.0 / Float(meridians))
      ½HexWidth = ¾HexWidth * 2.0 / 3.0
      radiusRatio = exquatorRadius / tubeRadius
   }
   
   public init(hexPairsWide: Int, hexColumnHeight: Int, ringRadius: Float) {
      self.meridians = hexPairsWide * 2
      self.½ParallelsPerHemisphere = hexColumnHeight
      self.exquatorRadius = ringRadius
      self.setConstants()
      
      createParallels()
      createHorizontalAxes()
      createTriangles()
   }

   private func createParallels() {
      Hexagon.owner = self
      
      parallels.append(Parallels(width: ¾HexWidth, latitude: Angle(inDegrees: 0.0), tubeRadius: tubeRadius))
      
      for p in 1...½ParallelsPerHemisphere {
         let width = parallels[p-1].longitude * (1.0 - distortion)
         let latitude = parallels[0].longitude * ½AspectRatio * radiusRatio * ((1.0 - pow(1.0 - distortion, Float(p))) / distortion)
         parallels.append(Parallels(width: width, latitude: latitude, tubeRadius: tubeRadius))
      }
      for p in 1..<½ParallelsPerHemisphere {
         let width = parallels[½ParallelsPerHemisphere-p].longitude
         let latitude = Angle(inDegrees: 360.0)-(parallels[½ParallelsPerHemisphere-p].latitude)
         parallels.append(Parallels(width: width, latitude: latitude, tubeRadius: tubeRadius))
      }
   }

   private func createHorizontalAxes() {
      for m in 0..<meridians {
         let parity = m % 2
         for p in stride(from: parity, to: (2 * ½ParallelsPerHemisphere), by: 2) {
            let (xLeft,zLeft,xRight,zRight) = horizontalAxisCoords(meridian: m, parallel: p)
            let ref = vertices.count
            vertices.append(Vertex(position: SIMD4<Float>(Float((xLeft + xRight)/2.0),
                                                          Float(parallels[p].yCoordinate),
                                                          Float((zLeft + zRight)/2.0), 1.0),
                                   colour:red))
            vertices.append(Vertex(position: SIMD4<Float>(Float(xLeft),
                                                          Float(parallels[p].yCoordinate),
                                                          Float(zLeft), 1.0),
                                   colour:yellow))
            vertices.append(Vertex(position: SIMD4<Float>(Float(xRight),
                                                          Float(parallels[p].yCoordinate),
                                                          Float(zRight), 1.0),
                                   colour:yellow))
            hexagons[m * meridians + p] = (Hexagon(meridian: m, parallel: p, centre: ref))
         }
      }
   }

   private func createTriangles() {
      Triangle.vertices = vertices
      for h in hexagons {
         indices.append(Triangle(x: h.value.centre, y: h.value.topLeft(), z: h.value.centreLeft))
         indices.append(Triangle(x: h.value.centre, y: h.value.topRight(), z: h.value.topLeft()))
         indices.append(Triangle(x: h.value.centre, y: h.value.centreRight, z: h.value.topRight()))
         indices.append(Triangle(x: h.value.centre, y: h.value.bottomRight(), z: h.value.centreRight))
         indices.append(Triangle(x: h.value.centre, y: h.value.bottomLeft(), z: h.value.bottomRight()))
         indices.append(Triangle(x: h.value.centre, y: h.value.centreLeft, z: h.value.bottomLeft()))
      }
   }
   
   public func printInitialisationReport() {
      let str0 = "Hex Tiled Torus built:\n"
      let str1 = String(format: "            Vertices: %8d                Indices: %8d\n", (vertices.count), (indices.count))
      let str2 = String(format: " Meridians (Columns): %8d       Parallels (Rows): %8d\n", meridians, ½ParallelsPerHemisphere)
      let str3 = String(format: "      Distortion/hex: %8.6f%%           AspectRatio: %8.6f%%\n", ((1.0 - pow(1.0 - distortion,2.0)) * 100), (½AspectRatio * 200.0))
      let str4 = String(format: "     Exquator Radius: %8.6f            Tube Radius: %8.6f\n", exquatorRadius, tubeRadius)
      let str5 = String(format: "     Inquator Radius: %8.6f             Rim Radius: %8.6f\n", (exquatorRadius - (tubeRadius * 2)), rimRadius)
      print(str0,str1,str2,str3,str4,str5)
   }
   
   public func listHexagons() {
      print("topLeft topRight  centreLeft      centre centreRight bottomLeft bottomRight")
      for h in hexagons.sorted(by: { $0.key < $1.key }) {
         print(h.value)
      }
   }
   
   public func listHexCentres() {
      print("\n\n(  m,  p )  ( longitude, latitude)      xCoord      yCoord       zCoord")
      for m in 0..<meridians {
         let parity = m % 2
         for p in stride(from: parity, to: (2 * ½ParallelsPerHemisphere), by: 2) {
            let xzAngle = Angle(inDegrees: (Float(m * 360)) / Float(meridians))
            let radius = (rimRadius + parallels[p].localRadius)
            let (x,z) = (cos(xzAngle.radians) * radius,
                         sin(xzAngle.radians) * radius)
            let strf = String(format: "( %2d, %2d )  (  %8.4f,%8.4f )    %8.4f    %8.4f     %8.4f     %8.4f",
                              arguments: [m, p, xzAngle.degrees, parallels[p].latitude.degrees, x, parallels[p].yCoordinate, z, radius])
            print(strf)
         }
         print()
      }
   }
   
   private func horizontalAxisCoords(meridian: Int, parallel: Int) -> (xLeftBound:Float, zLeftBound:Float, xRightBound:Float, zRightBound:Float) {
      // Use the hexagon centre point to calculate longitude angle and the radius of the torus ring at the height of the parallel
      let longitude = Angle(inDegrees: (Float(meridian * 360)) / Float(meridians))
      let radius = (rimRadius + parallels[parallel].localRadius)
      // Displace left and right to obtain horizontal axis end points
      let left = (longitude - ½HexWidth)
      let right = (longitude + ½HexWidth)
      return (cos(left.radians) * radius,
              sin(left.radians) * radius,
              cos(right.radians) * radius,
              sin(right.radians) * radius)
   }
   
   public var description: String {
      var str = "parallel   ¾ArcWidth    ½ArcHeight       radius      yCoord     radians"
      var index = 0
      
      for p in parallels {
         let myStr = "\(p)"
         let strf = String(format: "\n  %2d     %@",
                           arguments: [index, myStr])
         str += strf
         index += 1
      }
      return str
   }
}

private struct Parallels:CustomStringConvertible {
   let longitude: Angle
   let latitude: Angle
   let localRadius: Float
   let yCoordinate: Float
   
   init(width: Angle, latitude: Angle, tubeRadius: Float) {
      self.longitude = width
      self.latitude = latitude
      self.localRadius = cos(latitude.radians) * tubeRadius
      self.yCoordinate = sin(latitude.radians) * tubeRadius
   }
   
   var description: String {
      let strf = String(format: "  %8.4f      %8.4f    %8.4f      %8.4f",
                        arguments: [longitude.degrees, latitude.degrees, localRadius, yCoordinate])
      return strf
   }
}

public struct Triangle: CustomStringConvertible {
   static var vertices = [Vertex]()

   // Indices into vertices:[Vertex]
   public let x: UInt16
   public let y: UInt16
   public let z: UInt16
   public var faceNormal: SIMD4<Float> {
      let left = Triangle.vertices[Int(y)].position - Triangle.vertices[Int(x)].position
      let right = Triangle.vertices[Int(z)].position - Triangle.vertices[Int(x)].position
      return (left × right).normalised()
   }
   
   public var description: String {
      return String(format: "%4d, %4d, %4d,\n", x, y, z)
   }
}

fileprivate struct Hexagon: CustomStringConvertible {
   static var owner: HexagonTiledTorus!
   let meridian: Int // meridian index
   let parallel: Int // parallel index
   
   var belowParallel: Int { if parallel == 0 { return Hexagon.owner.½ParallelsPerHemisphere * 2 - 1 } else { return parallel - 1 }}
   var aboveParallel: Int { if parallel == (Hexagon.owner.½ParallelsPerHemisphere * 2 - 1) { return 0 } else { return parallel + 1 }}
   
   var previousMeridian: Int { if meridian == 0 { return Hexagon.owner.meridians - 1 } else { return meridian - 1 }}
   var nextMeridian: Int { if meridian == (Hexagon.owner.meridians - 1) { return 0 } else { return meridian + 1 }}
   
   // all the rest are indices into vertices array
   let centre: UInt16
   var centreLeft: UInt16
   var centreRight: UInt16
   
   init(meridian: Int, parallel: Int, centre: Int) {
      self.meridian = meridian
      self.parallel = parallel
      self.centre = UInt16(centre)
      self.centreLeft = UInt16(centre+1)
      self.centreRight = UInt16(centre+2)
   }
   
   func topLeft() -> UInt16 {
      return Hexagon.owner.hexagons[previousMeridian * Hexagon.owner.meridians + aboveParallel]!.centreRight
   }
   func topRight() -> UInt16 {
      return Hexagon.owner.hexagons[nextMeridian * Hexagon.owner.meridians + aboveParallel]!.centreLeft
   }
   func bottomLeft() -> UInt16 {
      return Hexagon.owner.hexagons[previousMeridian * Hexagon.owner.meridians + belowParallel]!.centreRight
   }
   func bottomRight() -> UInt16 {
      return Hexagon.owner.hexagons[nextMeridian * Hexagon.owner.meridians + belowParallel]!.centreLeft
   }
   
   var description: String {
      
      let strf = String(format: "\n  %4d     %4d        %4d        %4d        %4d       %4d        %4d",
                        arguments: [topLeft(), topRight(), centreLeft, centre, centreRight, bottomLeft(), bottomRight()])
      return strf
   }
}
