//
//  Renderer.swift
//  PlanetHex
//
//  Created by Don McBrien on 27/02/2020.
//  Copyright © 2020 Don McBrien. All rights reserved.
//

import Foundation
import MetalKit
import HexagonTiledTorus

class Renderer: NSObject, MTKViewDelegate {
   
   let model: Model
   let device: MTLDevice
   let commandQueue: MTLCommandQueue
   let renderPipelineState: MTLRenderPipelineState
   var vertexBuffer: MTLBuffer?
   var indexBuffer: MTLBuffer?
   var angleX = Float(0.0)
   var angleY = Float.pi / 3.0
   var angleZ = Float.pi / 7.0
   var depthStencilState: MTLDepthStencilState?
   
   func buildDepthStencilState() -> MTLDepthStencilState? {
     let descriptor = MTLDepthStencilDescriptor()
     descriptor.depthCompareFunction = .less
     descriptor.isDepthWriteEnabled = true
     return device.makeDepthStencilState(descriptor: descriptor)
   }
   
   init?(view: MTKView, model: Model) {
      self.model = model
      // First obtain and store the default device ...
      guard let defaultDevice = MTLCreateSystemDefaultDevice() else {
         fatalError("Metal is not supported on this device")
      }
      device = defaultDevice
      print("My GPU is: \(device)")
      
      // ... and attach the device to the view
      view.device = device
      view.depthStencilPixelFormat = .depth32Float

      // Then get an available command queue and store it
      guard let cQ = device.makeCommandQueue() else { return nil }
      commandQueue = cQ
      //library
      guard let library = device.makeDefaultLibrary() else {
         fatalError("library error")
      }

      //pipeline state
      do {
         let descriptor = MTLRenderPipelineDescriptor()
         descriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
         descriptor.vertexFunction = library.makeFunction(name: "vertex_main")
         descriptor.fragmentFunction = library.makeFunction(name: "fragment_main")
         descriptor.depthAttachmentPixelFormat = .depth32Float
         renderPipelineState = try device.makeRenderPipelineState(descriptor: descriptor)
      } catch let error as NSError {
         fatalError("PipelineState error: " + error.description)
      }
//      print(model.indices!,model.vertices!)
      indexBuffer = device.makeBuffer(bytes: model.indices,
                                      length: MemoryLayout<UInt16>.stride * model.indices.count,
                                      options: [])
      vertexBuffer = device.makeBuffer(bytes: model.vertices,
                                       length: MemoryLayout<Vertex>.stride * model.vertices.count,
                                       options: [])

      super.init()
      depthStencilState = buildDepthStencilState()!
   }
   
   func draw(in view: MTKView) {
      // Get the default MTLRenderPassDescriptor from the MTKView argument
      // in order to configure the render pass using it
      // set up render pass
      
      if let renderPassDescriptor = view.currentRenderPassDescriptor,
         let drawable = view.currentDrawable {
         
         // Change default settings. For example, we change the clear color from black to blue-green.
         let clearColor = MTLClearColor(red: 0.1, green: 0.1, blue: 0.1, alpha: 1)
         renderPassDescriptor.colorAttachments[0].clearColor = clearColor
         
         // Get an available command buffer
         //guard let commandBuffer = commandQueue.makeCommandBuffer() else { return }
         let commandBuffer = commandQueue.makeCommandBuffer()!
         
         // Get an available render command encoder
         guard let renderEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor) else { return }
         renderEncoder.setDepthStencilState(depthStencilState)
         
         angleX = angleX + 0.001
         angleY = angleY + 0.003
         angleZ = angleZ + 0.005

         var matrix = matrix_identity_float4x4

         var rotateX = matrix_identity_float4x4
         var rotateY = matrix_identity_float4x4
         var rotateZ = matrix_identity_float4x4
         // on x
         rotateX.columns.0 = [1, 0, 0, 0]
         rotateX.columns.1 = [0.0, cos(angleX), -sin(angleX), 0]
         rotateX.columns.2 = [0.0, sin(angleX), cos(angleX), 0]
         // on y
         rotateY.columns.0 = [cos(angleY), 0, sin(angleY), 0]
         rotateY.columns.1 = [0, 1, 0, 0]
         rotateY.columns.2 = [-sin(angleY), 0, cos(angleY), 0]
         // on z
         rotateZ.columns.0 = [cos(angleZ), -sin(angleZ), 0, 0]
         rotateZ.columns.1 = [sin(angleZ), cos(angleZ), 0, 0]
         rotateZ.columns.2 = [0, 0, 1, 0]

         var translate = matrix_identity_float4x4
         translate.columns.3 = [0, 0, 0.5, 1]

         var scale = matrix_identity_float4x4
//         let scaleX: Float = 1.0
//         let scaleY: Float = 1.0
//         let scaleZ: Float = 1.0
//         scale.columns.0 = [scaleX, 0, 0, 0]
//         scale.columns.1 = [0, scaleY, 0, 0]
//         scale.columns.2 = [0, 0, scaleZ, 0]

         matrix = translate * rotateZ * rotateY * rotateX * scale

         renderEncoder.setVertexBytes(&matrix,
                                      length: MemoryLayout<float4x4>.stride,
                                      index: 1)

         
         
         
         
         
         
         // Normally we would add some render commands here for encoding
         // Normally we would add some render commands here for encoding
         renderEncoder.setRenderPipelineState(renderPipelineState)
         
         
         // Comment out this to see the hexes
         renderEncoder.setTriangleFillMode(.lines)

         renderEncoder.setVertexBuffer(vertexBuffer, offset: 0, index: 0)

         renderEncoder.drawIndexedPrimitives(type: .triangle,
                                             indexCount: model.indices.count,
                                             indexType: .uint16,
                                             indexBuffer: indexBuffer!,
                                             indexBufferOffset: 0)

         // This finalizes the encoding of drawing commands and compiles them for the GPU.
         renderEncoder.endEncoding()
         
         // Tell Metal to send the rendering result to the MTKView when rendering completes
         commandBuffer.present(drawable)
         
         // Finally, send the encoded command buffer to the GPU.
         commandBuffer.commit()
      }
   }
   
   func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
      print("Window resized")
   }   
}

